# Generated by Django 3.2.15 on 2022-10-20 16:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0039_customerprofile_meta'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payment',
            name='provider',
            field=models.CharField(blank=True, max_length=30, null=True, verbose_name='Payment Provider'),
        ),
        migrations.AlterField(
            model_name='payment',
            name='status',
            field=models.IntegerField(choices=[(1, 'Queued'), (2, 'Active'), (10, 'Authorized'), (15, 'Captured'), (20, 'Settled'), (30, 'Canceled'), (35, 'Refunded'), (40, 'Declined'), (45, 'Error'), (50, 'Void')], default=0, verbose_name='Status'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='status',
            field=models.IntegerField(choices=[(10, 'Pause'), (20, 'Active'), (30, 'Canceled'), (40, 'Suspended'), (50, 'Expired')], default=0, verbose_name='Status'),
        ),
    ]



import uuid

from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from django.utils import timezone, dateformat

from vendor.models.base import CreateUpdateModelBase, SoftDeleteModelBase
from vendor.models.choice import SubscriptionStatus, PurchaseStatus
from vendor.utils import get_payment_scheduled_end_date


class SubscriptionReportModelManger(models.Manager):
    def get_total_cancelled_subscriptions(self, site, start_date=None, end_date=None):
        qs = super().get_queryset()
        
        if not (start_date and end_date):
            return qs.filter(profile__site=site, status=SubscriptionStatus.CANCELED)

        elif start_date and end_date is None:
            return qs.filter(profile__site=site, status=SubscriptionStatus.CANCELED, updated__date__gte=start_date)

        elif start_date is None and end_date:
            return qs.filter(profile__site=site, status=SubscriptionStatus.CANCELED, updated__date__lte=end_date)
            
        return qs.filter(profile__site=site, status=SubscriptionStatus.CANCELED, updated__date__gte=start_date, updated__date__lte=end_date)


class Subscription(SoftDeleteModelBase, CreateUpdateModelBase):
    '''
    A link for all the purchases a user has made. Contains subscription start and end date.
    This is generated for each item a user purchases so it can be checked in other code.
    '''
    uuid = models.UUIDField(_("UUID"), editable=False, unique=True, default=uuid.uuid4, null=False, blank=False)
    gateway_id = models.CharField(_("Subscription Gateway ID"), max_length=80)
    profile = models.ForeignKey("vendor.CustomerProfile", verbose_name=_("Purchase Profile"), on_delete=models.CASCADE, related_name="subscriptions")
    auto_renew = models.BooleanField(_("Auto Renew"), default=False)
    status = models.IntegerField(_("Status"), choices=SubscriptionStatus.choices, default=0)
    meta = models.JSONField(_("Meta"), default=dict, blank=True, null=True)

    reports = SubscriptionReportModelManger()

    class Meta:
        verbose_name = "Subscription"
        verbose_name_plural = "Subscriptions"

    def __str__(self):
        if self.receipts.count():
            return f"{self.receipts.first().order_item.name}"
        
        return f"{self.uuid}"

    @property
    def name(self):
        return self.__str__()

    def get_absolute_url(self):
        return reverse('vendor:customer-receipt', kwargs={'uuid': self.uuid})

    def void(self):
        """
        Funtion to void (not cancel) a receipt by making the end_date now and disabling
        access to the customer to that given product. If the receipt is related to a
        subscription to a Payment Gateway, make sure to also cancel such subscription
        in the given Payment Gateway.
        """
        active_receipts = self.receipts.filter(end_date__gte=timezone.now())

        for receipt in active_receipts:
            receipt.end_date = timezone.now()
            receipt.meta['voided_on'] = receipt.end_date.strftime("%Y-%m-%d_%H:%M:%S")
            receipt.save()
            
        self.meta[receipt.end_date.strftime("%Y-%m-%d_%H:%M:%S")] = f'voided receipst: {[receipt.uuid for receipt in active_receipts]}'
        self.save()

    def cancel(self):
        self.status = SubscriptionStatus.CANCELED
        self.auto_renew = False
        self.meta[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = 'Subscription Canceled'
        self.save()

    def get_next_billing_date(self):
        receipts = self.receipts.filter(Q(deleted=False), Q(end_date__gte=timezone.now()) | Q(end_date=None)).order_by('end_date')
        
        if not receipts.count():
            return None

        return receipts.first().end_date
        
    def get_last_payment_date(self):
        payment = self.payments.filter(deleted=False, status__lte=PurchaseStatus.SETTLED).order_by('-created').first()

        if not payment:
            return None

        return payment.get_receipt().start_date
        
    def is_on_trial(self):
        if not self.receipts.filter(deleted=False).count():
            return False

        receipt = self.receipts.filter(deleted=False).order_by('start_date').first()
        trial_end_date = receipt.start_date + timedelta(days=self.order_item.offer.get_trial_days())

        if timezone.now() > trial_end_date:
            return False
        
        return True

    def get_total(self):
        return self.receipts.first().order_item.total

    def save_payment_info(self, payment_info):
        if 'payment_info' not in self.meta:
            self.meta['payment_info'] = {}
        
        self.meta.update({'payment_info': payment_info})
        self.save()

    def get_offer(self):
        receipt = self.receipts.first()
        
        if not receipt:
            return None
        
        return receipt.order_item.offer





import uuid

from autoslug import AutoSlugField

from django.apps import apps
from django.conf import settings
from django.contrib.sites.models import Site
from django.contrib.sites.managers import CurrentSiteManager
from django.db import models
from django.utils.translation import gettext_lazy as _

from vendor.config import VENDOR_PRODUCT_MODEL, DEFAULT_CURRENCY

from .validator import validate_msrp
from .utils import is_currency_available
from .modelmanagers import SoftDeleteManager


def get_product_model():
    return apps.get_model(VENDOR_PRODUCT_MODEL)


##################
# DEFAULTS
##################
# TODO: Nice to have class MSRP(NestedModels)
def product_meta_default():
    return {'msrp': {'default': DEFAULT_CURRENCY, DEFAULT_CURRENCY: 0.00}}


def product_description_default():
    return {
        'description': '',
        'call_out': ' '
    }

def create_sku_if_empty(instance):
    return f"{instance.site.pk}-{instance.name}"

##################
# BASE MODELS
##################

class CreateUpdateModelBase(models.Model):
    '''
    This is a shared models base that provides created & updated timestamp fields
    '''
    created = models.DateTimeField("date created", auto_now_add=True)
    updated = models.DateTimeField("last updated", auto_now=True)

    class Meta:
        abstract = True


class SoftDeleteModelBase(models.Model):
    '''
    Adds "Soft Delete" functionality to models
    '''
    deleted = models.BooleanField(_("Deleted"), default=False)

    # Managers
    objects = models.Manager()                  # Standard Model manager
    not_deleted = SoftDeleteManager()                # Records filtered with deleted=False

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.deleted = True
        return self.save()


class ProductModelBase(CreateUpdateModelBase):
    '''
    This is the base class that all Products should inherit from.
    '''
    sku = models.CharField(_("SKU"), max_length=40, unique=True, blank=True, null=True, help_text=_("User Defineable SKU field"))   # Needs to be autogenerated by default, and unique from the PK
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)                                           # Used to track the product
    name = models.CharField(_("Name"), max_length=80, blank=False)
    site = models.ForeignKey(Site, verbose_name=_("Site"), on_delete=models.CASCADE, default=settings.SITE_ID, related_name="products")        # For multi-site support
    slug = AutoSlugField(populate_from='name', unique_with='site__id')                                                                         # Gets set in the save
    available = models.BooleanField(_("Available"), default=False, help_text=_("Is this currently available?"))        # This can be forced to be unavailable if there is no prices attached.
    description = models.JSONField(_("Description"), default=product_description_default, blank=True, null=True, help_text=_("Eg: {'call out': 'The ultimate product'}"))
    meta = models.JSONField(_("Meta"), validators=[validate_msrp], default=product_meta_default, blank=True, null=True, help_text=_("Eg: { 'msrp':{'usd':10.99} }\n(iso4217 Country Code):(MSRP Price)"))
    classification = models.ManyToManyField("vendor.TaxClassifier", blank=True)                                        # What taxes can apply to this item
    offers = models.ManyToManyField("vendor.Offer", blank=True, related_name="products")
    receipts = models.ManyToManyField("vendor.Receipt", blank=True, related_name="products")

    objects = models.Manager()
    on_site = CurrentSiteManager()

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def get_msrp(self, currency):
        if currency in self.meta['msrp']:
            return self.meta['msrp'][currency]
        else:
            return self.meta['msrp'][self.meta['msrp']['default']]

    def add_to_cart_url(self):
        """
        Link to add the item to the user's cart.
        """
    # TODO: ADD trigger when object becomes unavailable to disable offer if it exisits.

    def get_best_currency(self, currency=DEFAULT_CURRENCY):
        """
        If no currency is provided as an argument it will default to the products's msrp default currency.
        If currency is provided but is not available in the product it will default to the products's msrp default currency.
        """
        if is_currency_available(self.meta['msrp'].keys(), currency=currency):
            return currency
        else:
            return self.meta['msrp']['default']

    def active_profile_receipts(self):
        """
        Gets currently active reciepts by checking if the customer owns the product
        """
        return [receipt for receipt in self.receipts.all() if receipt.profile.has_product(self)]
    
    def owners(self):
        """
        Gets a set list of profiles that own the product
        """
        active_receipts = self.active_profile_receipts()
        return set([receipt.profile for receipt in active_receipts])

    def inactive_profile_receipts(self):
        """
        Gets a list of reciepts for customers that no longer own the product.
        """
        return [receipt for receipt in self.receipts.all() if not receipt.profile.has_product(self)]

    def expired_owners(self):
        """
        Gets a set list of profiles that no longer own the product
        """
        owners = self.owners()
        inactive_receipts = self.inactive_profile_receipts()
        return set([receipt.profile for receipt in inactive_receipts if receipt.profile not in owners])

    def save(self, *args, **kwargs):
        if not self.sku:
            self.sku = create_sku_if_empty(self)
        super().save(*args, **kwargs)



import json
import hashlib
import hmac
import logging

from django.conf import settings
from django.core.exceptions import PermissionDenied, MultipleObjectsReturned, ObjectDoesNotExist
from django.http import JsonResponse, HttpResponseRedirect
from django.views import View
from django.views.generic.edit import FormMixin
from django.views.decorators.csrf import csrf_exempt
from django.utils import timezone
from django.urls import reverse_lazy

from vendor.forms import DateTimeRangeForm
from vendor.integrations import AuthorizeNetIntegration
from vendor.models import Receipt, Invoice, Subscription, Payment
from vendor.models.choice import InvoiceStatus, PurchaseStatus
from vendor.processors.authorizenet import AuthorizeNetProcessor, sync_subscriptions
from vendor.utils import get_site_from_request

logger = logging.getLogger(__name__)


def update_payment(site, transaction_id, transaction_detail):
    processor = AuthorizeNetProcessor(site)

    submitted_datetime = datetime.strptime(transaction_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
    payment_info = processor.get_payment_info(transaction_detail)
    payment_status = processor.get_payment_status(transaction_detail.transactionStatus.text)
    payment_success = processor.get_payment_success(transaction_detail.responseCode.text)

    try:
        payment = Payment.objects.get(profile__site=site, transaction=transaction_id, status__lt=PurchaseStatus.VOID)
        payment.submitted_date = submitted_datetime
        payment.result[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = payment_info
        payment.status = payment_status
        payment.success = payment_success
        payment.save()
    except MultipleObjectsReturned as exce:
        logger.error(f"AuthorizeCaptureAPI update_payment multiple payments for transaction: {transaction_id} error: {exce}")

    except ObjectDoesNotExist as exce:
        logger.error(f"AuthorizeCaptureAPI update_payment payment does not exist for transaction: {transaction_id} error: {exce}")

    except Exception as exce:
        logger.error(f"AuthorizeCaptureAPI update_payment error: {exce}")
        
def subscription_save_transaction(site, transaction_id, transaction_detail):
    processor = AuthorizeNetProcessor(site)

    subscription_id = transaction_detail.subscription.id.text
    logger.info(f"AuthorizeCaptureAPI subscription_save_transaction saving subscription transaction: {transaction_id} for subscription {subscription_id}")

    submitted_datetime = datetime.strptime(transaction_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
    payment_info = processor.get_payment_info(transaction_detail)
    payment_status = processor.get_payment_status(transaction_detail.transactionStatus.text)
    payment_success = processor.get_payment_success(transaction_detail.responseCode.text)

    try:
        subscription = Subscription.objects.get(gateway_id=subscription_id, profile__site=site)

    except MultipleObjectsReturned as exce:
        logger.error(f"AuthorizeCaptureAPI subscription_save_transaction multiple subscription for id: {subscription_id} error: {exce}")
        subscription = Subscription.objects.filter(gateway_id=subscription_id, profile__site=site).first()

    except ObjectDoesNotExist as exce:
        logger.error(f"subscription_save_transaction subscription does not exist {subscription_id} exce: {exce}")
        return None

    try:
        payment = subscription.payments.get(transaction=None, status=PurchaseStatus.QUEUED)
        payment.transaction = transaction_id
        payment.submitted_date = submitted_datetime
        payment.result[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = payment_info
        payment.status = payment_status
        payment.success = payment_success
        payment.save()
        logger.info(f"AuthorizeCaptureAPI subscription_save_transaction: payment {payment.pk} updated")
        
        if payment_status == PurchaseStatus.SETTLED:
            processor = AuthorizeNetProcessor(site, payment.invoice)
            processor.payment = payment
            processor.create_receipts(payment.invoice.order_items.all())
            logger.info(f"AuthorizeCaptureAPI subscription_save_transaction: subscription renewed {subscription.pk}")

    except MultipleObjectsReturned as exce:
        # There should be none or only one payment with transaction None and status in Queue
        logger.error(f"AuthorizeCaptureAPI subscription_save_transaction multiple payments returned with None as Transaction, for {subscription_id} exce: {exce}")
        return None

    except ObjectDoesNotExist as exce:
        logger.info(f"AuthorizeCaptureAPI renew_subscription_task subscription {subscription.pk} renewed")
        invoice = Invoice.objects.create(
            profile=customer_profile,
            site=site,
            ordered_date=submitted_datetime,
            total=transaction_detail.settleAmount.pyval,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(offer)
        invoice.save()

        processor = AuthorizeNetProcessor(site, invoice)
        processor.subscription = subscription
        
        processor.renew_subscription(subscription, transaction_id, payment_status, payment_success)
        logger.info(f"AuthorizeCaptureAPI subscription_save_transaction creating new payment and receipt for subscription, for {subscription_id}")
        
        return None # No need to continue to create receipt as it is done in the above function

def settle_authorizenet_transactions(site, start_date, end_date):
    processor = AuthorizeNetProcessor(site)

    settled_transactions = processor.get_settled_transactions(start_date, end_date)
    processor.update_payments_to_settled(site, settled_transactions)


class AuthorizeNetBaseAPI(View):
    """
    Base class to handel Authroize.Net webhooks.
    """

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        """
        Dispatch override to accept post entries without csfr given that they include
        a X-Anet-Signature in there header that must be encoded with Signature Key using
        HMAC-sha512 on the payload.
        Reference: https://developer.authorize.net/api/reference/features/webhooks.html
        """
        return super().dispatch(*args, **kwargs)

    def is_valid_post(self, site):
        logger.info(f"AuthorizeNetBaseAPI is_valid_post: Request body: {self.request.body}")
        logger.info(f"AuthorizeNetBaseAPI is_valid_post: X ANET SIGNATURE: {self.request.META.get('HTTP_X_ANET_SIGNATURE')}")

        if 'HTTP_X_ANET_SIGNATURE' not in self.request.META:
            logger.warning("AuthorizeNetBaseAPI is_valid_post: Signature Key not it request")
            return False

        try:
            credentials = AuthorizeNetIntegration(site)

            if credentials.instance and credentials.instance.private_key:
                hash_value = hmac.new(bytes(credentials.instance.private_key, 'utf-8'), self.request.body, hashlib.sha512).hexdigest()
            elif settings.AUTHORIZE_NET_SIGNATURE_KEY:
                hash_value = hmac.new(bytes(settings.AUTHORIZE_NET_SIGNATURE_KEY, 'utf-8'), self.request.body, hashlib.sha512).hexdigest()
            else:
                raise TypeError("No private key set")

            logger.info(f"AuthorizeNetBaseAPI is_valid_post: Checking hashs\nCALCULATED: {hash_value}\nREQUEST VALUE: {self.request.META.get('HTTP_X_ANET_SIGNATURE')[7:]}")

            if hash_value.upper() == self.request.META.get('HTTP_X_ANET_SIGNATURE')[7:]:
                return True

        except TypeError as exce:
            logger.error(f'AuthorizeNetBaseAPI is_valid_post: TypeError Exception: {exce}')

        return False


class AuthorizeCaptureAPI(AuthorizeNetBaseAPI):
    """
    API endpoint to get event notifications from authorizenet when a authcaputre is created.
    If there is a subscription tied to the transaction, it will renew such subscription
    """
    def is_request_valid(self, site, request_data):
        if not request_data:
            logger.warning("AuthorizeCaptureAPI post: Webhook event has no body")
            return False

        if not request_data.get('payload', {}).get('id'):
            logger.error(f"AuthorizeCaptureAPI post: No transaction id request data: {request_data}")
            return False
            
        if not self.is_valid_post(site):
            logger.error(f"AuthorizeCaptureAPI post: Request was denied: {self.request}")
            return False

            
        return True

    def post(self, *args, **kwargs):
        logger.info(f"AuthorizeCaptureAPI post: Event webhook: {self.request.body}")

        site = get_site_from_request(self.request)
        logger.info(f"AuthorizeCaptureAPI post: site: {site}")

        request_data = json.loads(self.request.body)
        logger.info(f"AuthorizeCaptureAPI post: request data: {request_data}")

        if not self.is_request_valid(site, request_data):
            return JsonResponse({"msg": "AuthorizeCaptureAPI post: Request is invalid"})

        transaction_id = request_data.get('payload').get('id')
        logger.info(f"AuthorizeCaptureAPI post: Getting transaction detail for id: {transaction_id}")

        processor = AuthorizeNetProcessor(site)
        transaction_detail = processor.get_transaction_detail(transaction_id)

        if not hasattr(transaction_detail, 'subscription'):
            logger.info(f"AuthorizeCaptureAPI post: updating payment for transaction detail: {transaction_detail}")
            update_payment(site, transaction_id, transaction_detail)

        else:
            logger.info(f"AuthorizeCaptureAPI post: saving subscription transaction: {transaction_detail}")
            subscription_save_transaction(site, request_data.get('payload'), transaction_detail)

        return JsonResponse({"msg": "AuthorizeCaptureAPI post event finished"})


class VoidAPI(AuthorizeNetBaseAPI):

    def post(self, *args, **kwargs):
        logger.info(f"VoidAPI post: Event webhook: {self.request.body}")
        site = get_site_from_request(self.request)
        logger.info(f"VoidAPI post: site: {site}")

        if not self.request.body:
            logger.warning("VoidAPI post: Webhook event has no body")
            return JsonResponse({"msg": "VoidAPI post: Webhook event has no body"})

        if not self.is_valid_post(site):
            logger.error(f"VoidAPI post: Request was denied: {self.request}")
            raise PermissionDenied()
        
        request_data = json.loads(self.request.body)
        logger.info(f"VoidAPI post: request data: {request_data}")

        if request_data.get('eventType') != 'net.authorize.payment.void.created':
            logger.error(f"VoidAPI post: wrong event type: {request_data.get('eventType')}")
            return JsonResponse({"msg": "Event type is incorrect"})
        
        Payment.objects.filter(profile__site=site, transaction=request_data.get('payload').get('id')).update(status=PurchaseStatus.VOID)
        logger.info(f"VoidAPI post: payment with transaction voided: {request_data.get('payload').get('id')}")

        return JsonResponse({"msg": "VoidAPI post: success"})


class SyncSubscriptionsView(View):

    def get(self, *args, **kwargs):
        site = get_site_from_request(self.request)
        sync_subscriptions(site)
        return JsonResponse({'msg': 'Sync Subscriptions Finished'})


class GetSettledTransactionsView(FormMixin, View):
    form_class = DateTimeRangeForm
    success_url = reverse_lazy('vendor:vendor-home')

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(request.POST)
        site = get_site_from_request(self.request)

        if form.is_valid():
            settle_authorizenet_transactions(site, start_date, end_date)

        return HttpResponseRedirect(self.request.META.get('HTTP_REFERER', self.get_success_url()))


# Generated by Django 3.2.15 on 2022-09-01 12:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0038_auto_20220719_0944'),
    ]

    operations = [
        migrations.AddField(
            model_name='customerprofile',
            name='meta',
            field=models.JSONField(blank=True, default=dict, null=True, verbose_name='Meta'),
        ),
    ]



import uuid

from django.conf import settings
from django.contrib.sites.models import Site
from django.contrib.sites.managers import CurrentSiteManager
from django.db import models
from django.db.models import Q, QuerySet, Count, Sum
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from .base import CreateUpdateModelBase
from .choice import CURRENCY_CHOICES, TermType, PurchaseStatus, SubscriptionStatus
from .invoice import Invoice
from .utils import set_default_site_id
from vendor.config import DEFAULT_CURRENCY

from vendor.models.base import get_product_model
from vendor.models.choice import InvoiceStatus


#####################
# CUSTOMER PROFILE
#####################
class CustomerProfile(CreateUpdateModelBase):
    '''
    Additional customer information related to purchasing.
    This is what the Invoices are attached to.  This is abstracted from the user model directly do it can be mre flexible in the future.
    '''
    uuid = models.UUIDField(_("UUID"), editable=False, unique=True, default=uuid.uuid4, null=False, blank=False)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_("User"), on_delete=models.CASCADE, related_name="customer_profile")
    currency = models.CharField(_("Currency"), max_length=4, choices=CURRENCY_CHOICES, default=DEFAULT_CURRENCY)      # User's default currency
    site = models.ForeignKey(Site, verbose_name=_("Site"), on_delete=models.CASCADE, default=set_default_site_id, related_name="customer_profile")                      # For multi-site support
    meta = models.JSONField(_("Meta"), default=dict, blank=True, null=True)
        
    objects = models.Manager()
    on_site = CurrentSiteManager()

    class Meta:
        verbose_name = "Customer Profile"
        verbose_name_plural = "Customer Profiles"

    def __str__(self):
        if not self.user:
            return "New Customer Profile"
        return f"{self.user.username} - {self.site}"

    @property
    def email(self):
        return self.user.email

    def get_customer_profile_display(self):
        return str(self.user.username) + _("Customer Profile")

    def revert_invoice_to_cart(self):
        cart = self.invoices.get(status=InvoiceStatus.CHECKOUT)
        cart.status = InvoiceStatus.CART
        cart.save()

    def get_cart(self):
        if self.has_invoice_in_checkout():
            self.revert_invoice_to_cart()
    
        cart, created = self.invoices.get_or_create(status=InvoiceStatus.CART)
        return cart

    def get_checkout_cart(self):
        return self.invoices.filter(status=InvoiceStatus.CHECKOUT).first()

    def get_cart_or_checkout_cart(self):
        checkout_status = self.invoices.filter(status=InvoiceStatus.CHECKOUT, deleted=False).annotate(item_count=Count('order_items')).order_by('-item_count')
        cart_status = self.invoices.filter(status=InvoiceStatus.CART, deleted=False).annotate(item_count=Count('order_items')).order_by('-item_count')

        if checkout_status.count() > 1:  # There should only be one invoice in checkout status
            for invoice in checkout_status.all()[1:]:
                invoice.delete()
        
        if checkout_status:  # If there is an invoice in checkout status it there should be no invoices in cart status
            for invoice in cart_status.all():
                invoice.delete()

            return checkout_status.first()
        
        if not cart_status:  # There is no invoice in checkout or cart. Create a new one for user.
            cart, created = self.invoices.get_or_create(site=self.site, status=InvoiceStatus.CART)
            return cart

        if cart_status.count() > 1:  # There is more the one invoice in cart status. Remove all except one.
            for invoice in cart_status.all()[1:]:
                invoice.delete()

        return self.invoices.filter(status=InvoiceStatus.CART, deleted=False).first()

    def has_invoice_in_checkout(self):
        return bool(self.invoices.filter(status=InvoiceStatus.CHECKOUT).count())

    def filter_products(self, products):
        """
        returns the list of receipts that the user has a receipt for filtered by the products provided.
        """
        now = timezone.now()

        # Queryset or List of model records
        if isinstance(products, QuerySet) or isinstance(products, list):
            return self.receipts.filter(Q(products__in=products),
                                        Q(deleted=False),
                                        Q(start_date__lte=now) | Q(start_date=None),
                                        Q(end_date__gte=now) | Q(end_date=None))

        # Single model record
        return self.receipts.filter(Q(products=products),
                                    Q(deleted=False),
                                    Q(start_date__lte=now) | Q(start_date=None),
                                    Q(end_date__gte=now) | Q(end_date=None))

    def has_product(self, products):
        """
        returns true/false if the user has a receipt to a given product(s)
        it also checks against elegibility start/end/empty dates on consumable products and subscriptions
        """
        return bool(self.filter_products(products).count())

    def has_owned_product(self, products):
        # Queryset or List of model records
        if isinstance(products, QuerySet) or isinstance(products, list):
            return bool(self.receipts.filter(products__in=products).count())

        # Single model record
        return bool(self.receipts.filter(products=products).count())

    def get_recurring_receipts(self):
        """
        Gets the recurring receipts the customer profile might have
        """
        return self.receipts.filter(order_item__offer__terms__lt=TermType.PERPETUAL)

    def get_one_time_transaction_receipts(self):
        """
        Gets one time transation receipt the customer profile might have
        """
        return self.receipts.filter(order_item__offer__terms__gte=TermType.PERPETUAL)

    def get_cart_items_count(self):
        cart = self.get_cart_or_checkout_cart()

        return cart.order_items.all().count()

    def get_or_create_address(self, address):
        address, created = self.addresses.get_or_create(name=address.address_1, first_name=address.first_name, last_name=address.last_name, address_1=address.address_1, address_2=address.address_2, locality=address.locality, state=address.state, country=address.country, postal_code=address.postal_code, profile=self)
        return address, created

    def has_previously_owned_products(self, products):
        """
        Get all products that the customer has purchased and returns True if it has.
        """
        return bool(self.receipts.filter(products__in=products).first())

    def get_all_customer_products(self):
        Product = get_product_model()
        return Product.objects.filter(receipts__profile=self)

    def get_active_products(self):
         return set([receipt.products.first()  for receipt in self.get_active_receipts()])

    def get_active_offer_receipts(self, offer):
        return self.receipts.filter(Q(deleted=False), Q(order_item__offer=offer), Q(end_date__gte=timezone.now()) | Q(end_date=None))

    def get_active_receipts(self):
        return self.receipts.filter(Q(end_date__gte=timezone.now()) | Q(end_date=None))

    def get_inactive_receipts(self):
        return self.receipts.filter(end_date__lt=timezone.now())

    def get_active_product_and_offer(self):
        """
        Returns a tuple product and offer tuple that are related to the active receipt
        """
        return [(receipt.products.first(), receipt.order_item.offer) for receipt in self.receipts.filter(Q(deleted=False), Q(end_date__gte=timezone.now()) | Q(end_date=None))]

    def get_subscriptions(self):
        return self.subscriptions.all()

    def get_active_subscriptions(self):
        return self.subscriptions.filter(status=SubscriptionStatus.ACTIVE)

    def get_next_billing_date(self):
        next_billing_dates = []
        
        if not self.subscriptions.filter(status=SubscriptionStatus.ACTIVE).count():
            return None

        next_billing_dates = [subscription.get_next_billing_date() for subscription in self.get_active_subscriptions()]

        return sorted(next_billing_dates)[0]

    def get_last_payment_date(self):
        last_payment_dates = []
        
        if not self.subscriptions.count():
            return None

        last_payment_dates = [subscription.get_last_payment_date() for subscription in self.subscriptions.all()]

        return sorted(last_payment_dates)[-1]

    def get_payment_counts(self):
        return self.payments.filter(deleted=False, status=PurchaseStatus.SETTLED).count()

    def get_payment_sum(self):
        return self.payments.filter(deleted=False, status=PurchaseStatus.SETTLED).aggregate(Sum('amount'))

    def get_settled_payments(self):
        return self.payments.filter(deleted=False, status=PurchaseStatus.SETTLED).order_by('amount', 'submitted_date')

    def is_on_trial(self, offer):

        on_trial_receipt = next((receipt for receipt in self.get_active_offer_receipts(offer) if receipt.transaction and 'trial' in receipt.transaction), None)

        if on_trial_receipt:
            return True

        return False



import json
import logging
import stripe

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponse
from django.db.models import TextChoices
from django.utils import timezone
from django.utils.translation import gettext as _
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from vendor.models.choice import InvoiceStatus, PurchaseStatus
from vendor.integrations import StripeIntegration
from vendor.utils import get_site_from_request
from vendor.models import CustomerProfile, Subscription, Invoice
from vendor.processors import StripeProcessor


logger = logging.getLogger(__name__)



# TODO: Need to add more validation to function example:
# The lowest number can only 50 which transaltes to $0.50
# Probably should also added it to the processor as a static function
def convert_integer_to_float(number):
    number_string = str(number)

    return float(f"{number_string[:-2]}.{number_string[-2:]}")


class StripeEvents(TextChoices):
    INVOICE_PAID = 'invoice.paid', _('Invoice Paid')
    INVOICE_PAYMENT_FAILED = 'invoice.payment_failed', _('Invoice Payment Failed')
    INOVICE_PAYMENT_SUCCEEDED = 'invoice.payment_succeeded', _('Invoice Payment Succeeded')
    PAYMENT_INTENT_SUCCEDED = 'payment_intent.succeeded', _("Payment Succeeded")
    CHARGE_SUCCEEDED = 'charge.succeeded', _('Charge Succeeded')


class StripeBaseAPI(View):

    def __init__(self, **kwargs):
        self.stripe_event = None  # Variable used to store the webhooks event data.

        super().__init__(**kwargs)

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def is_valid_post(self, site):
        try:
            credentials = StripeIntegration(site)

            if credentials.instance and credentials.instance.private_key:
                self.event = stripe.Event.construct_from(json.loads(self.request.body), credentials.instance.private_key)
            elif settings.STRIPE_PUBLIC_KEY:
                self.event = stripe.Event.construct_from(json.loads(self.request.body), settings.STRIPE_PUBLIC_KEY)
            else:
                return False
            
        except ValueError as exce:
            return False
        except Exception as exce:
            return False

        return True
    
    def is_incoming_event_correct_and_recurring(self, event, desired_event):
        if event.type != desired_event:
            return False

        # This check is recuired to make sure that the event is related to a subscription.
        if self.event.data.object.billing_reason != 'subscription_cycle':
            return False

        return True
    
    def is_incoming_event_correct(self, event, desired_event):
        if event.type != desired_event:
            return False

        return True


class StripeSubscriptionInvoicePaid(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(self.request)

        if not self.is_valid_post(site):
            return HttpResponse(status=400)

        if not self.is_incoming_event_correct_and_recurring(self.event, StripeEvents.INVOICE_PAID):
            return HttpResponse(status=400)

        stripe_invoice = self.event.data.object
        paid_date = timezone.datetime.fromtimestamp(stripe_invoice.status_transitions.paid_at)
        amount_paid = convert_integer_to_float(stripe_invoice.total)

        # TODO nice to move this try/except blocks inside a generic function in the StripeBaseAPI class
        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400)

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(meta__stripe_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400)

        offer = subscription.get_offer()

        if not offer:
            logger.error(f"StripeSubscriptionInvoicePaid {subscription} has no offer attached")
            return HttpResponse(status=400)

        invoice = Invoice.objects.create(
            profile=customer_profile,
            site=site,
            ordered_date=paid_date,
            total=amount_paid,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(offer)
        invoice.vendor_notes = {'stripe_id': stripe_invoice.stripe_id}
        invoice.save()

        processor = StripeProcessor(site, invoice)
        processor.renew_subscription(subscription, stripe_invoice.charge, PurchaseStatus.SETTLED, payment_success=True)

        return HttpResponse(status=200)

class StripeSubscriptionPaymentFailed(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(request)

        if not self.is_valid_post(site):
            return HttpResponse(status=400)

        if not self.is_incoming_event_correct_and_recurring(self.event, StripeEvents.INVOICE_PAYMENT_FAILED):
            return HttpResponse(status=400)

        stripe_invoice = self.event.data.object
        paid_date = timezone.datetime.fromtimestamp(stripe_invoice.status_transitions.paid_at)

        # TODO nice to move this try/except blocks inside a generic function in the StripeBaseAPI class
        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400)

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(meta__stripe_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400)

        offer = subscription.get_offer()

        if not offer:
            logger.error(f"StripeSubscriptionInvoicePaid {subscription} has no offer attached")
            return HttpResponse(status=400)

        invoice = Invoice.objects.create(
            profile=subscription.profile,
            site=subscription.profile.site,
            ordered_date=paid_date,
            status=InvoiceStatus.COMPLETE
        )

        invoice.add_offer(offer)
        invoice.vendor_notes = {}
        invoice.vendor_notes['stripe_id'] = stripe_invoice.stripe_id
        invoice.save()

        processor = StripeProcessor(site, invoice)
        processor.subscription_payment_failed(subscription, stripe_invoice.charge)

        return HttpResponse(status=200)

class StripeInvoicePaid(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(self.request)

        if not self.is_valid_post(site):
            return HttpResponse(status=400)

        if not self.is_incoming_event_correct(self.event, StripeEvents.INOVICE_PAYMENT_SUCCEEDED):
            return HttpResponse(status=400)

        stripe_invoice = self.event.data.object

        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400)

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(gateway_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400)

        payment = subscription.payments.filter(transaction=None, status=PurchaseStatus.QUEUED).first()
        receipt = subscription.receipts.filter(transaction=None).first()

        if not payment and not receipt:
            logger.warning(f"There are no payments to update for subscription: {subscription}. Stripe Invoice: {stripe_invoice.id}")
            return HttpResponse(status=400)
        
        payment.transaction = stripe_invoice.charge
        payment.status = PurchaseStatus.SETTLED
        payment.save()

        receipt.transaction = stripe_invoice.charge
        receipt.save()

        return HttpResponse(status=200)


class StripeSyncObjects(View):

    def get(self, request, *args, **kwargs):
        site = get_site_from_request(request)

        processor = StripeProcessor(site)
        processor.sync_stripe_vendor_objects(site)

        return HttpResponse(status=200)



"""
Payment processor for Authorize.net.
"""
import ast
import logging

from datetime import datetime
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db.models import IntegerChoices
from django.utils import timezone
from math import ceil
from vendor.config import VENDOR_PAYMENT_PROCESSOR, VENDOR_STATE
from vendor.utils import get_future_date_days, get_payment_scheduled_end_date
from vendor.integrations import AuthorizeNetIntegration

logger = logging.getLogger(__name__)

try:
    from authorizenet import apicontractsv1
    from authorizenet import constants
    from authorizenet.apicontrollers import *
    import pyxb


    class CustomDate(pyxb.binding.datatypes.date):
        def __new__(cls, *args, **kw):
            # Because of some python, XsdLiteral (pyxb.binding.datatypes)
            # When a new date is created that is not a datetime and those, has more arguments,
            # it requires to only have the year, month and day arguments.

            if len(args) == 8:
                args = args[:3]
            return super().__new__(cls, *args, **kw)
            
except ModuleNotFoundError:
    if VENDOR_PAYMENT_PROCESSOR == "authorizenet.AuthorizeNetProcessor":
        print("WARNING: authorizenet module not found.  Install the library if you want to use the AuthorizeNetProcessor.")
        raise
    pass

from vendor.forms import CreditCardForm, BillingAddressForm
from vendor.models.choice import SubscriptionStatus, TransactionTypes, PaymentTypes, TermType, TermDetailUnits, InvoiceStatus, PurchaseStatus
from vendor.models import Invoice, Payment, Subscription, Receipt, Offer, CustomerProfile
from vendor.models.address import Country
from .base import PaymentProcessorBase


class AuthorizeNetProcessor(PaymentProcessorBase):
    """
    Implementation of AUTHORIZE.Net SDK
    https://apitest.authorize.net/xml/v1/schema/AnetApiSchema.xsd
    """

    AUTHORIZE = "authOnlyTransaction"
    AUTHORIZE_CAPTURE = "authCaptureTransaction"
    CAPTURE = "captureOnlyTransaction"
    REFUND = "refundTransaction"
    PRIOR_AUTHORIZE_CAPTURE = "priorAuthCaptureTransaction"
    VOID = "voidTransaction"
    GET_DETAILS = "getDetailsTransaction"
    AUTHORIZE_CONTINUE = "authOnlyContinueTransaction"
    AUTHORIZE_CAPTURE_CONTINUE = "authCaptureContinueTransaction"

    GET_SETTLED_BATCH_LIST = "getSettledBatchListRequest"

    """
    Controller executes the transaction, that process a transaction type
    """
    controller = None
    transaction = None
    merchant_auth = None
    transaction_type = None
    subscription_details = []

    def __str__(self):
        return 'Authorize.Net'


    def processor_setup(self, site):
        """
        Merchant Information needed to aprove the transaction.
        """
        self.merchant_auth = apicontractsv1.merchantAuthenticationType()
        self.credentials = AuthorizeNetIntegration(site)

        if self.credentials.instance:
            self.merchant_auth.name = self.credentials.instance.client_id
            self.merchant_auth.transactionKey = self.credentials.instance.public_key
        elif settings.AUTHORIZE_NET_TRANSACTION_KEY and settings.AUTHORIZE_NET_API_ID:
            self.merchant_auth.transactionKey = settings.AUTHORIZE_NET_TRANSACTION_KEY
            self.merchant_auth.name = settings.AUTHORIZE_NET_API_ID
        else:
            logger.error("AuthorizeNetProcessor Missing Authorize.net keys in settings: AUTHORIZE_NET_TRANSACTION_KEY and/or AUTHORIZE_NET_API_ID")
            raise ValueError("Missing Authorize.net keys in settings: AUTHORIZE_NET_TRANSACTION_KEY and/or AUTHORIZE_NET_API_ID")

        self.init_payment_type_switch()
        self.init_transaction_types()

    def init_transaction_types(self):
        self.transaction_types = {
            TransactionTypes.AUTHORIZE: self.AUTHORIZE,
            TransactionTypes.CAPTURE: self.CAPTURE,
            TransactionTypes.REFUND: self.REFUND,
            TransactionTypes.VOID: self.VOID,
        }

    def init_payment_type_switch(self):
        """
        Initializes the Payment Types create functions
        """
        self.payment_type_switch = {
            PaymentTypes.CREDIT_CARD: self.create_credit_card_payment,
            PaymentTypes.BANK_ACCOUNT: self.create_bank_account_payment,
            PaymentTypes.PAY_PAL: self.create_pay_pal_payment,
            PaymentTypes.MOBILE: self.create_mobile_payment,
        }

    def set_api_endpoint(self):
        """
        Sets the API endpoint for debugging or production.It is dependent on the VENDOR_STATE
        enviornment variable. Default value is DEBUG for the VENDOR_STATE
        """
        if VENDOR_STATE == 'DEBUG':
            self.API_ENDPOINT = constants.SANDBOX
        elif VENDOR_STATE == 'PRODUCTION':
            self.API_ENDPOINT = constants.PRODUCTION

    ##########
    # Authorize.net Object creations
    ##########
    def set_controller_api_endpoint(self):
        """
        Sets the endpoint for the controller to point to test or production.
        self.API_ENDPOINT is set on Processor Initialization
        """
        self.controller.setenvironment(self.API_ENDPOINT)

    def create_transaction(self):
        """
        This creates the main transaction to be processed.
        """
        transaction = apicontractsv1.createTransactionRequest()
        transaction.merchantAuthentication = self.merchant_auth
        transaction.refId = self.get_transaction_id()[:20]
        return transaction

    def create_transaction_type(self, trans_type):
        """
        Creates the transaction type instance with the amount
        """
        transaction_type = apicontractsv1.transactionRequestType()
        transaction_type.transactionType = trans_type
        transaction_type.currencyCode = 'USD'
        return transaction_type

    def create_credit_card_payment(self):
        """
        Creates and credit card payment type instance form the payment information set.
        """
        creditCard = apicontractsv1.creditCardType()
        creditCard.cardNumber = self.payment_info.data.get('card_number')
        creditCard.expirationDate = "-".join([self.payment_info.data.get('expire_year'), self.payment_info.data.get('expire_month')])
        creditCard.cardCode = str(self.payment_info.data.get('cvv_number'))
        return creditCard

    def create_bank_account_payment(self):
        raise NotImplementedError

    def create_pay_pal_payment(self):
        raise NotImplementedError

    def create_mobile_payment(self):
        raise NotImplementedError

    def create_authorize_payment(self):
        """
        Creates a payment instance acording to the billing information captured
        """
        payment = apicontractsv1.paymentType()
        payment.creditCard = self.payment_type_switch[int(
            self.payment_info.data.get('payment_type'))]()
        return payment

    def create_customer_data(self):
        customerData = apicontractsv1.customerDataType()
        customerData.type = "individual"
        customerData.id = str(self.invoice.profile.user.pk)
        customerData.email = self.invoice.profile.user.email
        return customerData

    def create_customer_data_recurring(self):
        customerData = apicontractsv1.customerType()
        customerData.type = "individual"
        customerData.id = str(self.invoice.profile.user.pk)
        customerData.email = self.invoice.profile.user.email
        return customerData

    def set_transaction_request_settings(self, settings):
        raise NotImplementedError

    def set_transaction_request_client_ip(self, client_ip):
        raise NotImplementedError

    def create_line_item(self, order_item):
        """
        Create a single line order_item (products) that is attached to a line order_item array
        """
        line_item = apicontractsv1.lineItemType()
        line_item.itemId = str(order_item.pk)
        line_item.name = order_item.name[:30]
        line_item.description = str(order_item.offer.description)[:250]
        line_item.quantity = str(order_item.quantity)
        line_item.unitPrice = str(order_item.price)

        return line_item

    def create_line_item_array(self, order_items):
        """
        Creates a list o line order_items(products) that are attached to the transaction type
        """
        line_items = apicontractsv1.ArrayOfLineItem()
        for order_item in order_items:
            line_items.lineItem.append(self.create_line_item(order_item))

        return line_items

    def create_tax(self, tax):
        raise NotImplementedError

    def create_shipping(self, shipping):
        raise NotImplementedError

    def create_billing_address(self, api_address_type):
        """
        Creates Billing address to improve security in transaction
        """
        billing_address = api_address_type
        billing_address.firstName = " ".join(self.payment_info.data.get('full_name', "").split(" ")[:-1])[:50]
        billing_address.lastName = (self.payment_info.data.get('full_name', "").split(" ")[-1])[:50]
        billing_address.company = self.billing_address.data.get('billing-company', "")[:50]
        address_lines = self.billing_address.data.get('billing-address_1', "")

        if self.billing_address.data['billing-address_2']:
            address_lines += f", {self.billing_address.data['billing-address_2']}"

        billing_address.address = address_lines
        billing_address.city = self.billing_address.data.get("billing-locality", "")[:40]
        billing_address.state = self.billing_address.data.get("billing-state", "")[:40]
        billing_address.zip = self.billing_address.data.get("billing-postal_code")[:20]
        country = Country(int(self.billing_address.data.get("billing-country")))
        billing_address.country = str(country.name)
        
        return billing_address

    def create_customer(self):
        raise NotImplementedError

    def create_order_type(self):
        """
        Create Authorize.Net OrderType to add invoice and descriptions to the transaction
        """
        order = apicontractsv1.orderType()
        order.invoiceNumber = str(self.invoice.pk)
        order.description = self.get_transaction_id()

        return order
    
    def get_interval_units(self, subscription):
        if subscription.offer.term_details.get('term_units', TermDetailUnits.MONTH) == TermDetailUnits.DAY:
            return apicontractsv1.ARBSubscriptionUnitEnum.days
        return apicontractsv1.ARBSubscriptionUnitEnum.months

    def create_payment_scheduale_interval_type(self, subscription, subscription_type):
        """
        subscription: Type: OrderItem
        Create an interval schedule with fixed months as units for period length.
        It calculates that start date depending on the term_units and trial_occurrences defined in the term_details.
        term_units can either be by day or by month. Start date is the first billing date of the subscriptions.
        Eg. for a 1 year 1 month free subscription:
            term_unit=20 (Month), trial_occurrences=1
            start_date = now + 1 month
        Eg. for a 7 day free 1 Month subscription:
            term_units=10 (Day), trial_occurrences=7
            start_date = now + 7 days
        """
        start_date = subscription.offer.get_term_start_date()

        payment_schedule = apicontractsv1.paymentScheduleType()
        payment_schedule.interval = apicontractsv1.paymentScheduleTypeInterval()
        payment_schedule.interval.unit = self.get_interval_units(subscription)

        payment_schedule.interval.length = subscription.offer.get_period_length()
        payment_schedule.totalOccurrences = subscription.offer.get_payment_occurrences()

        if self.invoice.profile.has_owned_product(subscription.offer.products.all()):
            payment_schedule.startDate = CustomDate(start_date)
            payment_schedule.trialOccurrences = 0
        else:
            payment_schedule.startDate = CustomDate(get_future_date_days(start_date, subscription.offer.get_trial_days()))
            payment_schedule.trialOccurrences = subscription.offer.get_trial_occurrences()
                
        return payment_schedule

    def create_transaction_order_information(self, invoice_number, description):
        order = apicontractsv1.orderType()
        order.invoiceNumber = invoice_number
        order.description = description
        return order

    def create_paging(self, limit, offset=1):
        paging = apicontractsv1.Paging()
        paging.limit = limit
        paging.offset = offset

        return paging

    ##########
    # Django-Vendor to Authoriaze.net data exchange functions
    ##########
    def get_payment_success(self, response_code):
        if response_code == "1":
            return True
        
        return False

    def get_payment_status(self, transaction_status):
        if transaction_status == 'authorizedPendingCapture':
            return PurchaseStatus.AUTHORIZED
        elif transaction_status == 'capturedPendingSettlement':
            return PurchaseStatus.CAPTURED
        elif transaction_status == 'communicationError':
            return PurchaseStatus.ERROR
        elif transaction_status == 'refundSettledSuccessfully':
            return PurchaseStatus.REFUNDED
        elif transaction_status == 'approvedReview':
            return PurchaseStatus.AUTHORIZED
        elif transaction_status == 'declined':
            return PurchaseStatus.DECLINED
        elif transaction_status == 'couldNotVoid':
            return PurchaseStatus.ERROR
        elif transaction_status == 'expired':
            return PurchaseStatus.DECLINED
        elif transaction_status == 'generalError':
            return PurchaseStatus.ERROR
        elif transaction_status == 'failedReview':
            return PurchaseStatus.ERROR
        elif transaction_status == 'settledSuccessfully':
            return PurchaseStatus.SETTLED
        elif transaction_status == 'settlementError':
            return PurchaseStatus.ERROR
        elif transaction_status == 'underReview':
            return PurchaseStatus.ERROR
        elif transaction_status == 'voided':
            return PurchaseStatus.VOID
        elif transaction_status == 'FDSPendingReview':
            return PurchaseStatus.ERROR
        elif transaction_status == 'FDSAuthorizedPendingReview':
            return PurchaseStatus.ERROR
        elif transaction_status == 'returnedItem':
            return PurchaseStatus.REFUNDED
        else:
            raise TypeError(f"{transaction_status} is not supported, check Authorize.Net docs for transactionStatus field choices")

    def get_payment_info(self, transaction):
        account_number = transaction.payment.creditCard.cardNumber.text[-4:]
        full_name = " ".join([transaction.billTo.firstName.text, transaction.billTo.lastName.text])

        payment_info = super().get_payment_info(account_number, full_name)
        
        return payment_info.update({
            'account_type': transaction.payment.creditCard.cardType.text,
            'transaction_id': transaction.transId.text,
            'subscription_id': transaction.subscription.id.text if hasattr(transaction, 'subscription') else "-",
            'payment_number': transaction.subscription.payNum.text,
            'status': transaction.transactionStatus.text
        })

    def subscription_info_to_dict(self, subscription_info):
        return {
            'name': subscription_info.subscription.name.text,
            'interval': subscription_info.subscription.paymentSchedule.interval.length.text,
            'unit': subscription_info.subscription.paymentSchedule.interval.unit.text,
            'start_date': subscription_info.subscription.paymentSchedule.startDate.text,
            'total_occurrences': subscription_info.subscription.paymentSchedule.totalOccurrences.text,
            'trial_occurrences': subscription_info.subscription.paymentSchedule.trialOccurrences.text,
            'amount': subscription_info.subscription.amount.text,
            'trial_amount': subscription_info.subscription.trialAmount.text,
        }

    def get_vendor_subscription_status(self, subscription_status):
        if subscription_status == 'active':
            return SubscriptionStatus.ACTIVE
        elif subscription_status == 'expired':
            return SubscriptionStatus.EXPIRED
        elif subscription_status == 'suspended':
            return SubscriptionStatus.SUSPENDED
        elif subscription_status == 'canceled':
            return SubscriptionStatus.CANCELED
        elif subscription_status == 'terminated':
            return SubscriptionStatus.SUSPENDED
        else:
            raise TypeError(f"{subscription_status} status is not valid, take a look at Authorize.Net documentation")

    def get_transaction_raw_response(self):
        """
        Returns a dictionary with raw information about the current transaction
        """
        response = self.transaction_response.__dict__

        if 'errors' in response:
            response.pop('errors')

        if 'messages' in response:
            response.pop('messages')
            
        return str({**self.transaction_info, **response})

    def get_transaction_errors(self, transaction_data=None):
        errors = []

        if self.transaction_response.messages.resultCode == "Error":
            errors.append({key: value.text for key, value in self.transaction_response.messages.message.__dict__.items() if value.text})

        if transaction_data and hasattr(transaction_data, 'errors'):
            errors.append({key: value.text for key, value in transaction_data.errors.error.__dict__.items() if value.text})
        
        return errors

    def get_transaction_data(self, transaction_data=None):
        response_data = {key: value.text for key, value in self.transaction_response.__dict__.items() if value.text}

        if transaction_data:
            for key, value in transaction_data.__dict__.items():
                if value.text:
                    response_data.update({key: value.text})

        if 'messages' in self.transaction_response.__dict__:
            response_data['resultCode'] = self.transaction_response.messages.resultCode.text
            response_data['messages'] = {key: value.text for key, value in self.transaction_response.messages.message.__dict__.items() if value.text}

        return response_data

    def parse_response(self, parser_function):
        """
        Processes the transaction response from the gateway so it can be saved in the payment model
        """
        parser_function()

    def is_transaction_response_empty(self):

        if not self.transaction_response:
            self.set_transaction_info(
                raw="No transaction Response was set",
                errors="No transaction Response was set",
                messages="No transaction Response was set"
            )
            return True
        
        return False

    def parse_payment_response(self):
        self.transaction_info = {}

        if self.is_transaction_response_empty():
            return None

        errors = self.get_transaction_errors(self.transaction_response.transactionResponse)
        
        if not hasattr(self.transaction_response, 'transactionResponse'):
            self.set_transaction_info(
                raw=self.get_transaction_raw_response(),
                errors=errors
            )
            return None

        self.transaction_info = self.get_transaction_info(
            raw=self.get_transaction_raw_response(),
            errors=errors,
            data=self.get_transaction_data(self.transaction_response.transactionResponse)
        )

    def parse_transaction_response(self):
        self.transaction_info = {}

        if self.is_transaction_response_empty():
            return None

        errors = self.get_transaction_errors()

        self.transaction_info = self.get_transaction_info(
            raw=self.get_transaction_raw_response(),
            errors=errors,
            data=self.get_transaction_data()
        )

    def parse_success(self):
        self.transaction_succeeded = False

        if hasattr(self.transaction_response, 'messages') and\
           self.transaction_response.messages.resultCode == "Ok" and\
           not self.transaction_info['errors']:
            self.transaction_succeeded = True

    ##########
    # Base Processor Transaction Implementations
    ##########
    def process_payment(self):
        # Init transaction
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(settings.AUTHORIZE_NET_TRANSACTION_TYPE_DEFAULT)
        self.transaction_type.amount = self.to_valid_decimal(self.invoice.get_one_time_transaction_total())
        self.transaction_type.payment = self.create_authorize_payment()
        self.transaction_type.customer = self.create_customer_data()
        self.transaction_type.billTo = self.create_billing_address(apicontractsv1.customerAddressType())

        # Optional items for make it easier to read and use on the Authorize.net portal.
        if self.invoice.order_items:
            self.transaction_type.lineItems = self.create_line_item_array(self.invoice.order_items.all())

        # You set the request to the transaction
        self.transaction.transactionRequest = self.transaction_type
        self.controller = createTransactionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        # You execute and get the response
        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_payment_response)
        self.parse_success()

        self.transaction_id = self.transaction_info['data'].get('transId', "")

        if self.transaction_succeeded:
            self.payment.status = PurchaseStatus.CAPTURED
        else:
            self.payment.status = PurchaseStatus.DECLINED

        self.payment.save()

    def subscription_payment(self, subscription):
        """
        subscription: Type: OrderItem
        Creates a subscription for a user. Subscriptions can be monthy or yearly.objects.all()
        """
        # Setting subscription details
        self.transaction_type = apicontractsv1.ARBSubscriptionType()
        self.transaction_type.name = subscription.offer.name
        self.transaction_type.paymentSchedule = self.create_payment_scheduale_interval_type(subscription, subscription.offer.terms)
        self.transaction_type.amount = self.to_valid_decimal(subscription.total - subscription.discounts)
        self.transaction_type.trialAmount = self.to_valid_decimal(subscription.offer.get_trial_amount())
        self.transaction_type.billTo = self.create_billing_address(apicontractsv1.nameAndAddressType())
        self.transaction_type.payment = self.create_authorize_payment()
        self.transaction_type.customer = self.create_customer_data_recurring()

        # Optional to add Order information.
        self.transaction_type.order = self.create_order_type()

        # Creating the request
        self.transaction = apicontractsv1.ARBCreateSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscription = self.transaction_type

        # Creating and executing the controller
        self.controller = ARBCreateSubscriptionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        # Getting the response
        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        if self.transaction_succeeded:
            self.subscription_id = self.transaction_info['data'].get('subscription_id', "")

    def subscription_update_payment(self, subscription):
        """
        Updates the credit card information for the subscriptions in authorize.net
        and updates the payment record associated with the receipt.
        """
        self.transaction_type = apicontractsv1.ARBSubscriptionType()
        self.transaction_type.payment = self.create_authorize_payment()

        self.transaction = apicontractsv1.ARBUpdateSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = subscription.gateway_id
        self.transaction.subscription = self.transaction_type

        self.controller = ARBUpdateSubscriptionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        subscription_info = self.subscription_info(subscription.gateway_id)

        payment_info = {}
        account_number = getattr(subscription_info['subscription']['profile']['paymentProfile']['payment']['creditCard'], 'cardNumber', None)
        account_type = getattr(subscription_info['subscription']['profile']['paymentProfile']['payment']['creditCard'], 'accountType', None)

        if account_number:
            payment_info['account_number'] = account_number.text

        if account_type:
            payment_info['account_type'] = account_type.text

        subscription.save_payment_info(payment_info)

    def subscription_cancel(self, subscription):
        """
        If receipt.invoice.total is zero, no need to call Gateway as there is no
        transaction for it. Otherwise it will cancel the subscription on the Gateway
        and if successfull it will cancel it on the receipt.
        """
        super().subscription_cancel(subscription)
        
        self.transaction = apicontractsv1.ARBCancelSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = str(subscription.gateway_id)
        self.transaction.includeTransactions = False

        self.controller = ARBCancelSubscriptionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()
            
    def subscription_info(self, subscription_id):
        self.transaction = apicontractsv1.ARBGetSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = str(subscription_id)
        self.transaction.includeTransactions = True

        self.controller = ARBGetSubscriptionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()
        
        if self.transaction_succeeded:
            return self.transaction_response

        return None

    def refund_payment(self, payment):
        # Init transaction
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(
            self.transaction_types[TransactionTypes.REFUND])
        self.transaction_type.amount = self.to_valid_decimal(payment.amount)
        self.transaction_type.refTransId = payment.transaction

        creditCard = apicontractsv1.creditCardType()
        creditCard.cardNumber = ast.literal_eval(payment.result['payment_info']).get('account_number')[-4:]
        creditCard.expirationDate = "XXXX"

        payment_type = apicontractsv1.paymentType()
        payment_type.creditCard = creditCard
        self.transaction_type.payment = payment_type

        self.transaction.transactionRequest = self.transaction_type
        self.controller = createTransactionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_payment_response)
        self.parse_success()
        
        if self.transaction_succeeded:
            payment.status = PurchaseStatus.REFUNDED
            payment.save()

    def void_payment(self, transaction_id):
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(self.transaction_types[TransactionTypes.VOID])
        self.transaction_type.refTransId = transaction_id

        self.transaction.transactionRequest = self.transaction_type

        self.controller = createTransactionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_payment_response)
        self.parse_success()
        
        super().void_payment()

    def is_card_valid(self):
        """
        Handles an Authorize Only transaction to ensure that the funds are in the customers bank account
        """
        invoice_number = str(self.invoice.pk)[:19]
        description = "This amount is only to check for valid cards and will not be charged. Depending on your bank the charge can take 3 to 5 days to be removed."
        self.create_payment_model(settings.VENDOR_CHARGE_VALIDATION_PRICE)
        self.transaction = self.create_transaction()
        self.transaction_type = self.create_transaction_type(self.transaction_types[TransactionTypes.AUTHORIZE])
        self.transaction_type.amount = self.to_valid_decimal(settings.VENDOR_CHARGE_VALIDATION_PRICE)
        self.transaction_type.payment = self.create_authorize_payment()
        self.transaction_type.billTo = self.create_billing_address(apicontractsv1.customerAddressType())
        self.transaction_type.order = self.create_transaction_order_information(invoice_number, description)
        
        self.transaction.transactionRequest = self.transaction_type
        self.controller = createTransactionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        # You execute and get the response
        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_payment_response)
        self.parse_success()
        
        if self.transaction_succeeded:
            self.payment.transaction = self.transaction_info['data'].get('trans_id', "")
            self.payment.save()
            self.void_payment(self.payment.transaction)

            return True

        return False

    def subscription_update_price(self, subscription, new_price, user):
        self.transaction_type = apicontractsv1.ARBSubscriptionType()
        self.transaction_type.amount = self.to_valid_decimal(new_price)

        self.transaction = apicontractsv1.ARBUpdateSubscriptionRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.subscriptionId = str(subscription.gateway_id)
        self.transaction.subscription = self.transaction_type

        self.controller = ARBUpdateSubscriptionController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        if self.transaction_succeeded:
            super().subscription_update_price(subscription, new_price, user)

    def get_customer_id_for_expiring_cards(self, month):
        paging = apicontractsv1.Paging()
        paging.limit = 10
        paging.offset = 1

        sorting = apicontractsv1.CustomerPaymentProfileSorting()
        sorting.orderBy = apicontractsv1.CustomerPaymentProfileOrderFieldEnum.id
        sorting.orderDescending = "false"

        self.transaction = apicontractsv1.getCustomerPaymentProfileListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth

        self.transaction.searchType = apicontractsv1.CustomerPaymentProfileSearchTypeEnum.cardsExpiringInMonth
        self.transaction.month = month
        self.transaction.sorting = sorting
        self.transaction.paging = paging

        self.controller = getCustomerPaymentProfileListController(self.transaction)
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        customer_profile_ids = []
        last_page = 1
        
        if self.transaction_succeeded and self.transaction_response.paymentProfiles:
            last_page = ceil(self.transaction_response.totalNumInResultSet.pyval / paging.limit)
            customer_profile_ids.extend([customer_profile.customerProfileId.text for customer_profile in self.transaction_response.paymentProfiles.paymentProfile])

        for previous_page in range(1, last_page):
            paging.offset = previous_page + 1

            self.transaction.paging = paging
            self.controller = getCustomerPaymentProfileListController(self.transaction)
            self.controller.execute()
            self.transaction_response = self.controller.getresponse()
            self.parse_response(self.parse_transaction_response)
            self.parse_success()

            if self.transaction_succeeded and self.transaction_response.paymentProfiles:
                customer_profile_ids.extend([customer_profile.customerProfileId.text for customer_profile in self.transaction_response.paymentProfiles.paymentProfile])

        return customer_profile_ids
    
    def get_customer_email(self, customer_id):
        self.transaction = apicontractsv1.getCustomerProfileRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.customerProfileId = customer_id

        self.controller = getCustomerProfileController(self.transaction)
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        if self.transaction_succeeded:
            return self.transaction_response.profile.email.pyval
        
        return None

    def get_settled_transactions(self, start_date, end_date):
        batch_list = self.get_settled_batch_list(start_date, end_date)
        successfull_transactions = []

        if not batch_list:
            return []
        
        for batch in batch_list:
            transaction_list = self.get_transaction_batch_list(str(batch.batchId))
            successfull_transactions.extend([ transaction for transaction in transaction_list if transaction['transactionStatus'] == 'settledSuccessfully' ])

        return successfull_transactions
    
    def update_payments_to_settled(self, site, settled_transactions):
        for settled_transaction in settled_transactions:
            try:
                payment = Payment.objects.get(profile__site=site, transaction=settled_transaction.transId.text)
                payment.status = PurchaseStatus.SETTLED
                payment.submitted_date = settled_transaction.submitTimeUTC.pyval
                payment.save()
            except ObjectDoesNotExist as exce:
                logger.error(f"update_payments_to_settled payment for transaction: {settled_transaction.transId.text} was not found for site: {site}")


    ##########
    # Reporting API, for transaction retrieval information
    ##########
    def get_settled_batch_list(self, start_date, end_date):
        """
        Gets a list of batches for settled transaction between the start and end date.
        """
        self.transaction = apicontractsv1.getSettledBatchListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.firstSettlementDate = start_date
        self.transaction.lastSettlementDate = end_date

        self.controller = getSettledBatchListController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        if self.transaction_succeeded and hasattr(self.transaction_response, 'batchList'):
            return [batch for batch in self.transaction_response.batchList.batch]

    def get_transaction_batch_list(self, batch_id):
        """
        Gets the list of settled transaction in a batch. There are sorting and paging option
        that are not currently implemented. It will get the last 1k transactions.
        """
        self.transaction = apicontractsv1.getTransactionListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.batchId = batch_id

        self.controller = getTransactionListController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        response = self.controller.getresponse()

        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok and hasattr(response, 'transactions'):
            return [transaction for transaction in response.transactions.transaction]

    def get_transaction_detail(self, transaction_id):
        self.transaction = apicontractsv1.getTransactionDetailsRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.transId = transaction_id

        self.controller = getTransactionDetailsController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        response = self.controller.getresponse()

        if response.messages.resultCode == apicontractsv1.messageTypeEnum.Ok:
            return response.transaction

    def get_list_of_subscriptions(self, limit=1000, offset=1, search_type=apicontractsv1.ARBGetSubscriptionListSearchTypeEnum.subscriptionActive):
        self.transaction = apicontractsv1.ARBGetSubscriptionListRequest()
        self.transaction.merchantAuthentication = self.merchant_auth
        self.transaction.searchType = search_type
        self.transaction.paging = self.create_paging(limit, offset)

        self.controller = ARBGetSubscriptionListController(self.transaction)
        self.set_controller_api_endpoint()
        self.controller.execute()

        # Work on the response
        self.transaction_response = self.controller.getresponse()
        self.parse_response(self.parse_transaction_response)
        self.parse_success()

        if self.transaction_succeeded:
            return self.transaction_response.subscriptionDetails.subscriptionDetail
        else:
            return []

    def get_subscription_transactions(self, subscription_info):
        subscription_transactions = []
        if not hasattr(subscription_info.subscription, 'arbTransactions'):
            return []

        for transaction in subscription_info.subscription.arbTransactions.arbTransaction:
            if hasattr(transaction, 'transId'):
                transaction_detail = self.get_transaction_detail(transaction.transId.text)
                subscription_transactions.append(transaction_detail)

        return subscription_transactions


def sync_subscriptions(site):
    logger.info("sync_subscriptions Starting Subscription Migration")
    processor = AuthorizeNetProcessor(site)

    active_subscriptions = processor.get_list_of_subscriptions()
    subscription_ids = [ subscription.id.text for subscription in active_subscriptions ]

    inactive_subscriptions = processor.get_list_of_subscriptions(search_type=apicontractsv1.ARBGetSubscriptionListSearchTypeEnum.subscriptionInactive)
    subscription_ids.extend([ subscription.id.text for subscription in inactive_subscriptions ])
    
    for subscription_id in subscription_ids:
        subscription_info = processor.subscription_info(subscription_id)

        if hasattr(subscription_info.subscription.profile, 'email'):
            email = subscription_info.subscription.profile.email.text

            try:
                customer_profile = CustomerProfile.objects.get(site=site, user__email__iexact=email)
                offers = Offer.objects.filter(site=site, name=subscription_info.subscription.name)

                if not offers.count():
                    raise ObjectDoesNotExist()
                
                offer = offers.first()

                ## Create a subscription with status.
                subscription, _ = Subscription.objects.get_or_create(gateway_id=subscription_id, profile=customer_profile)
                subscription.status = processor.get_vendor_subscription_status(subscription_info.subscription.status.text)
                subscription.meta = processor.subscription_info_to_dict(subscription_info)
                subscription.save()

                ## Get transactions for subscription.
                subscription_transactions = processor.get_subscription_transactions(subscription_info)

                for transaction in subscription_transactions:
                    transaction_id = transaction.transId.text
                    transaction_detail = processor.get_transaction_detail(transaction_id)
                    
                    submitted_datetime = datetime.strptime(transaction_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
                    payment_info = processor.get_payment_info(transaction_detail)
                    payment_status = processor.get_payment_status(transaction_detail.transactionStatus.text)
                    payment_success = processor.get_payment_success(transaction_detail.responseCode.text)

                    if not subscription.payments.filter(transaction=transaction_id).count():
                        ### Create Invoice
                        invoice = Invoice.objects.create(
                            profile=customer_profile,
                            site=site,
                            ordered_date=submitted_datetime,
                            total=transaction_detail.settleAmount.pyval,
                            status=InvoiceStatus.COMPLETE
                        )
                        invoice.add_offer(offer)
                        invoice.save()

                        ### Create Payment
                        payment = Payment()
                        payment.profile = customer_profile
                        payment.invoice = invoice
                        payment.subscription = subscription
                        payment.amount = transaction_detail.settleAmount.pyval
                        payment.submitted_date = submitted_datetime
                        payment.status = payment_status
                        payment.success = payment_success
                        payment.result = {}
                        payment.result['payment_info'] = payment_info
                        payment.transaction = transaction_id
                        payment.payee_full_name = payment_info['full_name']
                        payment.save()

                        if payment_status == PurchaseStatus.SETTLED:
                            ### Create Receipt.
                            receipt = Receipt()
                            receipt.profile = customer_profile
                            receipt.order_item = invoice.order_items.first()
                            receipt.start_date = submitted_datetime
                            receipt.end_date = get_payment_scheduled_end_date(offer, start_date=submitted_datetime)
                            receipt.transaction = payment.transaction
                            receipt.subscription = subscription
                            receipt.meta.update(payment.result)
                            receipt.meta['payment_amount'] = payment.amount
                            receipt.save()
                            
                            receipt.products.add(offer.products.first())

                    else:
                        # Update the Payment
                        payment = subscription.payments.get(transaction=transaction_id)
                        payment.amount = transaction_detail.settleAmount.pyval
                        payment.submitted_date = submitted_datetime
                        payment.status = payment_status
                        payment.success = payment_success
                        payment.result = {}
                        payment.result['payment_info'] = payment_info
                        payment.payee_full_name = payment_info['full_name']
                        payment.save()

                        if not payment.invoice:                        ### Create Invoice
                            invoice = Invoice.objects.create(
                                profile=customer_profile,
                                site=site,
                                ordered_date=submitted_datetime,
                                total=transaction_detail.settleAmount.pyval,
                                status=InvoiceStatus.COMPLETE
                            )
                            invoice.add_offer(offer)
                            invoice.save()
                            
                            payment.invoice = invoice
                            payment.save()
                        else:
                            invoice = payment.invoice

                        if Receipt.objects.filter(deleted=False, transaction=transaction_id, profile=customer_profile).count():
                            receipt = Receipt.objects.get(deleted=False, transaction=transaction_id, profile=customer_profile)
                            
                            if payment.status == PurchaseStatus.SETTLED:
                                receipt.profile = customer_profile
                                receipt.order_item = invoice.order_items.first()
                                receipt.start_date = submitted_datetime
                                receipt.end_date = get_payment_scheduled_end_date(offer, start_date=submitted_datetime)
                                receipt.transaction = payment.transaction
                                receipt.subscription = subscription
                                receipt.meta.update(payment.result)
                                receipt.meta['payment_amount'] = payment.amount
                                receipt.save()
                                
                                receipt.products.add(offer.products.first())
                            else:
                                # Only receipts for settled transactions should exists
                                receipt.deleted = True
                                receipt.save()

            except ObjectDoesNotExist as exce:
                logger.error(f"sync_subscriptions exception: {exce}")
            except MultipleObjectsReturned as exce:
                logger.error(f"sync_subscriptions exception: {exce}")
            except Exception as exce:
                logger.error(f"sync_subscriptions exception: {exce}")

def sync_subscriptions_and_create_missing_receipts(site):
    logger.info("sync_subscriptions_and_create_missing_receipts Starting Subscription Migration")
    processor = AuthorizeNetProcessor(site)
    
    subscriptions = processor.get_list_of_subscriptions(1000)
    active_subscription_ids = [ subscription.id.text for subscription in subscriptions if subscription['status'] == 'active' ]
    
    for subscription_id in active_subscription_ids:
        try:
            subscription = Subscription.objects.get(gateway_id=subscription_id, profile__site=site)

        except ObjectDoesNotExist as exce:
            pass

        except Exception as exce:
            pass

        subscription_info = processor.subscription_info(subscription_id)
        # valid_subscription_transactions = [transaction for transaction in subscription_info.subscription.arbTransactions.arbTransaction if hasattr(transaction, 'transId') ]
        valid_subscription_transactions = []
        if hasattr(subscription_info.subscription, 'arbTransactions'):
            for transaction in subscription_info.subscription.arbTransactions.arbTransaction:
                if hasattr(transaction, 'transId'):
                    valid_subscription_transactions.append(transaction)

        for transaction in valid_subscription_transactions:
            logger.info(f"sync_subscriptions_and_create_missing_receipts Processing transaction {transaction}")

            transaction_id = transaction.transId.text

            if not subscription.payments.filter(transaction=transaction_id).count():
                trans_processor = AuthorizeNetProcessor(site)

                trans_detail = trans_processor.get_transaction_detail(transaction_id)
                submitted_datetime = datetime.strptime(trans_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
                offer = Offer.objects.get(site=site, name=subscription_info.subscription.name)
                
                invoice = Invoice.objects.create(
                    status=InvoiceStatus.COMPLETE,
                    site=site,
                    profile=subscription.profile,
                    ordered_date=submitted_datetime,
                    total=trans_detail.settleAmount.pyval
                )
                invoice.add_offer(offer)
                invoice.save()

                payment_info = {
                    'account_number': trans_detail.payment.creditCard.cardNumber.text[-4:],
                    'account_type': trans_detail.payment.creditCard.cardType.text,
                    'full_name': " ".join([trans_detail.billTo.firstName.text, trans_detail.billTo.lastName.text]),
                    'transaction_id': transaction_id,
                    'subscription_id': trans_detail.subscription.id.text,
                    'payment_number': trans_detail.subscription.payNum.text
                }

                # Create Payment
                payment = Payment(profile=invoice.profile,
                            amount=invoice.total,
                            invoice=invoice
                            )
                payment.result = {}
                payment.result['payment_info'] = payment_info
                payment.subscription = subscription
                payment.success = True
                payment.status = PurchaseStatus.SETTLED
                payment.transaction = transaction_id
                payment.payee_full_name = payment_info['full_name']
                payment.amount = trans_detail.settleAmount.pyval
                payment.submitted_date = trans_detail.submitTimeUTC.pyval
                payment.save()

                # Create Receipt
                receipt = Receipt()
                receipt.profile = invoice.profile
                receipt.order_item = invoice.order_items.first()
                receipt.transaction = payment.transaction
                receipt.meta.update(payment.result)
                receipt.meta['payment_amount'] = payment.amount
                receipt.start_date = submitted_datetime
                receipt.save()

                receipt.products.add(offer.products.first())
                receipt.end_date = get_payment_scheduled_end_date(offer, start_date=receipt.start_date)
                receipt.subscription = subscription
                receipt.save()



import logging

from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from stripe import Customer
from vendor.models import CustomerProfile, Offer
from vendor.config import PaymentProcessorSiteConfig, SupportedPaymentProcessor, DEFAULT_CURRENCY
from vendor.processors import StripeProcessor, StripeQueryBuilder


logger = logging.getLogger(__name__)

@receiver(post_save, sender=CustomerProfile)
def stripe_create_customer_signal(sender, instance, created, **kwargs):
    site_configured_processor = PaymentProcessorSiteConfig(instance.site)

    if site_configured_processor.get_key_value('payment_processor') != SupportedPaymentProcessor.STRIPE.value:
        logger.error(f"stripe_create_customer_signal instance: {instance.pk} was not created on stripe for site {instance.site} because site is not configured with Stripe")
        return None
    
    if 'stripe_id' in instance.meta:
        return None


    processor = StripeProcessor(instance.site)

    query_builder = StripeQueryBuilder()

    email_clause = query_builder.make_clause_template(
        field='email',
        value=instance.user.email,
        operator=query_builder.EXACT_MATCH,
        next_operator=query_builder.AND
    )

    metadata_clause = query_builder.make_clause_template(
        field='metadata',
        key='site',
        value=instance.site.domain,
        operator=query_builder.EXACT_MATCH
    )

    query = query_builder.build_search_query(processor.stripe.Customer, [email_clause, metadata_clause])
    query_result = processor.stripe_query_object(processor.stripe.Customer, query)

    if not processor.transaction_succeeded:
        logger.error(f"stripe_create_customer_signal: {processor.transactions_info}")
        return None 

    if query_result:
        logger.info(f"stripe_create_customer_signal: updating customer")
        processor.update_stripe_customers([instance])
        if len(query_result.data) > 1:
            logger.warning(f"stripe_create_customer_signal: more than one customer found for email: {instance.user.email} on site: {instance.site.domain}")
        return None

    processor.create_stripe_customers([instance])
    logger.info(f"stripe_create_customer_signal instance: {instance.pk} successfully created on Stripe")


@receiver(post_delete, sender=CustomerProfile)
def stripe_delete_customer_signal(sender, instance, **kwargs):
    site_configured_processor = PaymentProcessorSiteConfig(instance.site)

    if site_configured_processor.get_key_value('payment_processor') != SupportedPaymentProcessor.STRIPE.value:
        logger.error(f"stripe_delete_customer_signal instance: {instance.pk} can't be deleted as this site: {instance.site} is not configured to use Stripe")
        return None
    
    if 'stripe_id' not in instance.meta:
        return None

    processor = StripeProcessor(instance.site)

    processor.stripe_delete_object(processor.stripe.Customer, instance.meta['stripe_id'])
    logger.info(f"stripe_delete_customer_signal instance: {instance.pk} was successfully deleted on Stripe")


@receiver(post_save, sender=Offer)
def stripe_create_offer_signal(sender, instance, created, **kwargs):
    site_configured_processor = PaymentProcessorSiteConfig(instance.site)

    if site_configured_processor.get_key_value('payment_processor') != SupportedPaymentProcessor.STRIPE.value:
        logger.error(
            f"stripe_create_offer_signal instance: {instance.pk} was not created on stripe for site {instance.site} because site is not configured with Stripe")
        return None

    processor = StripeProcessor(instance.site)

    query_builder = StripeQueryBuilder()

    pk_clause = query_builder.make_clause_template(
        field='metadata',
        key='pk',
        value=str(instance.pk),
        operator=query_builder.EXACT_MATCH,
        next_operator=query_builder.AND
    )

    metadata_clause = query_builder.make_clause_template(
        field='metadata',
        key='site',
        value=instance.site.domain,
        operator=query_builder.EXACT_MATCH
    )

    query = query_builder.build_search_query(processor.stripe.Product, [pk_clause, metadata_clause])
    query_result = processor.stripe_query_object(processor.stripe.Product, query)

    if not processor.transaction_succeeded:
        logger.error(f"stripe_create_offer_signal: {processor.transactions_info}")
        return None

    if query_result['data']:
        logger.info(f"stripe_create_offer_signal: updating the offer")
        processor.update_offers([instance])
        return None

    processor.create_offers([instance])
    logger.info(f"stripe_create_offer_signal instance: offer {instance.pk} successfully created on Stripe")


@receiver(post_delete, sender=Offer)
def stripe_delete_offer_signal(sender, instance, **kwargs):
    site_configured_processor = PaymentProcessorSiteConfig(instance.site)

    if site_configured_processor.get_key_value('payment_processor') != SupportedPaymentProcessor.STRIPE.value:
        logger.error(
            f"stripe_delete_offer_signal instance: {instance.pk} can't be deleted as this site: {instance.site} is not configured to use Stripe")
        return None

    if 'stripe' not in instance.meta:
        return None

    processor = StripeProcessor(instance.site)

    processor.stripe_delete_object(processor.stripe.Product, instance.meta['stripe']['product_id'])
    processor.stripe_update_object(processor.stripe.Price, instance.meta['stripe']['price_id'], {'active': False})

    if instance.meta.get('stripe').get('coupon_id'):
        processor.stripe_delete_object(processor.stripe.Coupon, instance.meta['stripe']['coupon_id'])
        logger.info(f"stripe_delete_offer_signal instance: offer {instance.pk} coupon was successfully deleted on Stripe")

    logger.info(f"stripe_delete_offer_signal instance: offer {instance.pk} was successfully deleted on Stripe")



from django.urls import path

from vendor.api.v1 import views as api_views
from vendor.api.v1.authorizenet import views as authorizenet_views
from vendor.api.v1.stripe import views as stripe_views

app_name = "vendor_api"

urlpatterns = [
    path('', api_views.VendorIndexAPI.as_view(), name='api-index'),
    path('cart/add/<slug:slug>/', api_views.AddToCartView.as_view(), name="add-to-cart"),
    path('cart/remove/<slug:slug>/', api_views.RemoveFromCartView.as_view(), name="remove-from-cart"),
    path('customer/subscription/<uuid:uuid>/cancel/', api_views.SubscriptionCancelView.as_view(), name="customer-subscription-cancel"),
    path('product/<uuid:uuid>/remove', api_views.VoidProductView.as_view(), name="manager-profile-remove-product"),
    path('profile/<uuid:uuid_profile>/offer/<uuid:uuid_offer>/add', api_views.AddOfferToProfileView.as_view(), name="manager-profile-add-offer"),
    path('product/<uuid:uuid>/availability', api_views.ProductAvailabilityToggleView.as_view(), name="manager-product-availablility"),
    path('subscription/price/update', api_views.SubscriptionPriceUpdateView.as_view(), name="manager-subscription-price-update"),
    # AuthorizeNet
    path('authorizenet/authcapture', authorizenet_views.AuthorizeCaptureAPI.as_view(), name='api-authorizenet-authcapture'),
    path('authorizenet/void', authorizenet_views.VoidAPI.as_view(), name='api-authorizenet-void'),
    path('authorizenet/sync/subscriptions/', authorizenet_views.SyncSubscriptionsView.as_view(), name='api-authorizenet-sync-subscriptions'),
    path('authorizenet/settled/transactions/', authorizenet_views.GetSettledTransactionsView.as_view(), name='api-authorizenet-settled-transactions'),
    # Stripe
    path('stripe/invoice/paid/', stripe_views.StripeInvoicePaid.as_view(), name='api-stripe-invoice-paid'),
    path('stripe/subscription/invoice/paid/', stripe_views.StripeSubscriptionInvoicePaid.as_view(), name='api-stripe-subscription-invoice-paid'),
    path('stripe/subscription/invoice/payment/failed/', stripe_views.StripeSubscriptionPaymentFailed.as_view(), name='api-stripe-subscription-invoice-payment-failed'),
    path('stripe/sync/objects/', stripe_views.StripeSyncObjects.as_view(), name='api-stripe-sync-objects'),
]


# Generated by Django 3.2.13 on 2022-06-01 17:45

from django.db import migrations


def pre_invoice_status_change(apps, schema_editor):
    InvoiceModel = apps.get_model('vendor', 'Invoice')

    print(f"Invoice Status Updating {InvoiceModel.objects.filter(status__gt=10).count()}")
    InvoiceModel.objects.filter(status__gt=10).update(status=20)
    print(f"Invoice Status Updated {InvoiceModel.objects.filter(status__gt=10).count()}")


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0030_auto_20220414_1055'),
    ]

    operations = [
        migrations.RunPython(pre_invoice_status_change, reverse_code=migrations.RunPython.noop),
    ]



from calendar import mdays
from datetime import timedelta, datetime, timezone
from decimal import Decimal, ROUND_UP
from django.contrib.sites.shortcuts import get_current_site
from django.utils import timezone

from vendor.models.choice import TermDetailUnits


#############
# Requst and Session Utils
def get_site_from_request(request):
    if hasattr(request, 'site'):
        return request.site
    return get_current_site(request)

def get_or_create_session_cart(session):
    session_cart = {}
    if 'session_cart' not in session:
        session['session_cart'] = session_cart
    session_cart = session.get('session_cart')

    return session_cart

def clear_session_purchase_data(request):
    if 'billing_address_form' in request.session:
        del(request.session['billing_address_form'])
    if 'credit_card_form' in request.session:
        del(request.session['credit_card_form'])


def get_future_date_months(today, add_months):
    """
    Returns a datetime object with the a new added months
    """
    newday = today.day
    newmonth = (((today.month - 1) + add_months) % 12) + 1
    newyear = today.year + (((today.month - 1) + add_months) // 12)
    if newday > mdays[newmonth]:
        newday = mdays[newmonth]
    if newyear % 4 == 0 and newmonth == 2:
        newday += 1
    return datetime(newyear, newmonth, newday, tzinfo=timezone.utc)


def get_future_date_days(today, add_days):
    """
    Returns a datetime object with the a new added days
    """
    return today + timedelta(days=add_days)


def get_payment_scheduled_end_date(offer, start_date=timezone.now()):
    """
    Determines the start date offset so the payment gateway starts charging the monthly offer
    """
    units = offer.term_details.get('term_units', TermDetailUnits.MONTH)

    if units == TermDetailUnits.MONTH:
        return get_future_date_months(start_date, offer.get_period_length())
    elif units == TermDetailUnits.DAY:
        return get_future_date_days(start_date, offer.get_period_length())

def get_subscription_start_date(offer, profile, start_date=timezone.now()):
    if profile.is_on_trial(offer):
        return start_date + timedelta(days=offer.get_trial_days())
    
    if not profile.has_owned_product(offer.products.all()):
        return start_date + timedelta(days=offer.get_trial_days())

    return start_date
        

def get_display_decimal(amount):
    return Decimal(amount).quantize(Decimal('.00'), rounding=ROUND_UP)





from vendor.models.utils import generate_sku
from vendor.models.validator import validate_msrp_format, validate_msrp

from vendor.models.base import ProductModelBase

from vendor.models.address import Address
from vendor.models.invoice import Invoice, OrderItem
from vendor.models.offer import Offer, offer_term_details_default
from vendor.models.payment import Payment
from vendor.models.price import Price
from vendor.models.profile import CustomerProfile
from vendor.models.receipt import Receipt
from vendor.models.subscription import Subscription
from vendor.models.tax import TaxClassifier
from vendor.models.wishlist import Wishlist, WishlistItem
# from .product import Product
